""":mod:`wirerope` --- Universal method/function wrapper.
=========================================================
"""

# this line must be placed first for setup.cfg
from ._version import __version__
from .wire import Wire
from .rope import WireRope, RopeCore

__all__ = ('__version__', 'Wire', 'RopeCore', 'WireRope')



""":mod:`wirerope.rope` --- Wire access dispatcher for descriptor type.
=======================================================================
"""
import six
from .callable import Callable
from .wire import descriptor_bind
from ._compat import functools

__all__ = 'WireRope', 'RopeCore'


class RopeCore(object):
    """The base rope object.

    To change rope behavior, create a subclass or compatible class
    and pass it to `core_class` argument of :class wirerope.rope.WireRope`.
    """

    def __init__(self, callable, rope):
        super(RopeCore, self).__init__()
        self.callable = callable
        self.rope = rope

    @property
    def wire_class(self):
        return self.rope.wire_class


class MethodRopeMixin(object):

    def __init__(self, *args, **kwargs):
        super(MethodRopeMixin, self).__init__(*args, **kwargs)
        assert not self.callable.is_barefunction

    def __set_name__(self, owner, name):
        # Use a non-identifier character as separator to prevent name clash.
        self.wire_name = '|'.join(['__wire', owner.__name__, name])

    def __get__(self, obj, type=None):
        cw = self.callable
        co = cw.wrapped_object
        owner, _ = descriptor_bind(co, obj, type)
        if owner is None:  # invalid binding but still wire it
            owner = obj if obj is not None else type
        if hasattr(self, 'wire_name'):
            wire_name = self.wire_name
            # Lookup in `__dict__` instead of using `getattr`, because
            # `getattr` falls back to class attributes.
            wire = owner.__dict__.get(wire_name)
        else:
            wire_name_parts = ['__wire_', cw.wrapped_callable.__name__]
            if owner is type:
                wire_name_parts.extend(('_', type.__name__))
            wire_name = ''.join(wire_name_parts)
            wire = getattr(owner, wire_name, None)
        if wire is None:
            wire = self.wire_class(self, owner, (obj, type))
            setattr(owner, wire_name, wire)
        assert callable(wire.__func__)
        return wire


class PropertyRopeMixin(object):

    def __init__(self, *args, **kwargs):
        super(PropertyRopeMixin, self).__init__(*args, **kwargs)
        assert not self.callable.is_barefunction

    def __set_name__(self, owner, name):
        # Use a non-identifier character as separator to prevent name clash.
        self.wire_name = '|'.join(['__wire', owner.__name__, name])

    def __get__(self, obj, type=None):
        cw = self.callable
        co = cw.wrapped_object
        owner, _ = descriptor_bind(co, obj, type)
        if owner is None:  # invalid binding but still wire it
            owner = obj if obj is not None else type
        if hasattr(self, 'wire_name'):
            wire_name = self.wire_name
            # Lookup in `__dict__` instead of using `getattr`, because
            # `getattr` falls back to class attributes.
            wire = owner.__dict__.get(wire_name)
        else:
            wire_name_parts = ['__wire_', cw.wrapped_callable.__name__]
            if owner is type:
                wire_name_parts.extend(('_', type.__name__))
            wire_name = ''.join(wire_name_parts)
            wire = getattr(owner, wire_name, None)
        if wire is None:
            wire = self.wire_class(self, owner, (obj, type))
            setattr(owner, wire_name, wire)
        return wire._on_property()  # requires property path


class FunctionRopeMixin(object):

    def __init__(self, *args, **kwargs):
        super(FunctionRopeMixin, self).__init__(*args, **kwargs)
        assert self.callable.is_barefunction or self.callable.is_boundmethod
        self._wire = self.wire_class(self, None, None)

    def __getattr__(self, name):
        try:
            return self.__getattribute__(name)
        except AttributeError:
            pass
        return getattr(self._wire, name)


class CallableRopeMixin(object):

    def __init__(self, *args, **kwargs):
        super(CallableRopeMixin, self).__init__(*args, **kwargs)
        self.__call__ = functools.wraps(self.callable.wrapped_object)(self)

    def __call__(self, *args, **kwargs):
        return self._wire(*args, **kwargs)


class WireRope(object):
    """The end-user wrapper for callables.

    Any callable can be wrapped by this class regardless of its concept -
    free function, method, property or even more weird one.
    The calling type is decided by each call and redirected to proper
    RopeMixin.

    The rope will detect method or property owner by needs. It also will return
    or call their associated `wirerope.wire.Wire` object - which are the user
    defined behavior.
    """

    def __init__(
            self, wire_class, core_class=RopeCore,
            wraps=False, rope_args=None):
        self.wire_class = wire_class
        self.method_rope = type(
            '_MethodRope', (MethodRopeMixin, core_class), {})
        self.property_rope = type(
            '_PropertyRope', (PropertyRopeMixin, core_class), {})
        self.function_rope = type(
            '_FunctionRope', (FunctionRopeMixin, core_class), {})
        self.callable_function_rope = type(
            '_CallableFunctionRope',
            (CallableRopeMixin, FunctionRopeMixin, core_class), {})
        for rope in (self, self.method_rope, self.property_rope,
                     self.function_rope, self.callable_function_rope):
            rope._wrapped = wraps
            rope._args = rope_args

    def __call__(self, function):
        """Wrap a function/method definition.

        :return: Rope object. The return type is up to given callable is
                 function, method or property.
        """
        cw = Callable(function)
        if cw.is_barefunction or cw.is_boundmethod:
            rope_class = self.callable_function_rope
            wire_class_call = self.wire_class.__call__
            if six.PY3:
                if wire_class_call.__qualname__ == 'type.__call__':
                    rope_class = self.function_rope
            else:
                # method-wrapper test for CPython2.7
                # im_class == type test for PyPy2.7
                if type(wire_class_call).__name__ == 'method-wrapper' or \
                        wire_class_call.im_class == type:
                    rope_class = self.function_rope
        elif cw.is_property:
            rope_class = self.property_rope
        else:
            rope_class = self.method_rope
        rope = rope_class(cw, rope=self)
        if rope._wrapped:
            rope = functools.wraps(cw.wrapped_callable)(rope)
        return rope



from wirerope.callable import Callable


def test_callable_attributes():

    def f():
        pass
    w = Callable(f)
    assert w.is_barefunction is True
    assert w.is_boundmethod is False
    assert w.is_descriptor is False
    assert w.is_membermethod is False
    assert w.is_boundmethod is False
    assert w.is_classmethod is False
    assert w.is_property is False

    class A():

        def m(self):
            pass
        w = Callable(m)
        assert w.is_barefunction is False
        assert w.is_boundmethod is False
        assert w.is_descriptor is False
        assert w.is_membermethod is True
        assert w.is_boundmethod is False
        assert w.is_classmethod is False
        assert w.is_property is False

        @classmethod
        def c(cls):
            pass
        w = Callable(c)
        assert w.is_barefunction is False
        assert w.is_boundmethod is False
        assert w.is_descriptor is True
        assert w.is_membermethod is False
        assert w.is_boundmethod is False
        assert w.is_classmethod is True
        assert w.is_property is False

        @staticmethod
        def s():
            pass
        w = Callable(s)
        assert w.is_barefunction is False
        assert w.is_boundmethod is False
        assert w.is_descriptor is True
        assert w.is_membermethod is False
        assert w.is_boundmethod is False
        assert w.is_classmethod is False
        assert w.is_property is False

        @property
        def p(self):
            pass
        w = Callable(p)
        assert w.is_barefunction is False
        assert w.is_boundmethod is False
        assert w.is_descriptor is True
        assert w.is_membermethod is False
        assert w.is_boundmethod is False
        assert w.is_classmethod is False
        assert w.is_property is True

    w = Callable(A.m)
    assert w.is_barefunction is False
    assert w.is_boundmethod is False
    assert w.is_descriptor is False
    assert w.is_membermethod is True
    assert w.is_boundmethod is False
    assert w.is_classmethod is False
    assert w.is_property is False

    a = A()

    w = Callable(a.m)
    assert w.is_barefunction is False
    assert w.is_boundmethod is True
    assert w.is_descriptor is False
    assert w.is_membermethod is False
    assert w.is_boundmethod is True
    assert w.is_classmethod is False
    assert w.is_property is False

    w = Callable(a.c)
    assert w.is_barefunction is False
    assert w.is_boundmethod is True
    assert w.is_descriptor is False
    assert w.is_membermethod is False
    assert w.is_boundmethod is True
    assert w.is_classmethod is False
    assert w.is_property is False

    w = Callable(a.s)
    # assert w.is_barefunction is False
    assert w.is_descriptor is False
    assert w.is_property is False
    assert w.is_boundmethod is False
    assert w.is_classmethod is False
    # assert w.is_membermethod is False



from __future__ import absolute_import

import types
import six
from ._util import cached_property
from ._compat import inspect

__all__ = ('Callable', )


_inspect_iscoroutinefunction = getattr(
    inspect, 'iscoroutinefunction', lambda f: False)


class _Reagent(object):
    pass


_reagent = _Reagent()


def _f(owner):
    return owner


def _name_binder(descriptor, obj, type):
    return type, None


def _type_binder(descriptor, obj, type):
    return type, type


def _obj_binder(descriptor, obj, type):
    return obj, obj


_descriptor_binder_cache = {}


class Descriptor(object):

    def __init__(self, descriptor):
        self.descriptor = descriptor
        self.descriptor_class = type(descriptor)

    def detect_function_attr_name(self):
        indicator = object()
        descriptor = self.descriptor_class(indicator)
        for name in dir(descriptor):
            try:
                attr = getattr(descriptor, name)
            except AttributeError:
                continue
            if attr is indicator:
                # detected!
                return name
        else:
            raise RuntimeError(
                "The given function doesn't hold the given function as an "
                "attribute. Is it a correct descriptor?")

    def detect_property(self, obj, type_):
        d = self.descriptor_class(_f)
        method_or_value = d.__get__(obj, type_)
        return method_or_value is obj or method_or_value is type_

    def detect_binder(self, obj, type_):
        key = (self.descriptor_class, obj is not None)
        if key not in _descriptor_binder_cache:
            d = self.descriptor_class(_f)
            method = d.__get__(obj, type_)
            if isinstance(method, types.FunctionType):
                # not a boundmethod - probably staticmethod
                binder = _name_binder
            elif method is obj:
                binder = _obj_binder
            elif method is type_:
                binder = _type_binder
            elif callable(method):
                owner = method()
                if owner is type_:
                    binder = _type_binder
                elif owner is obj:
                    binder = _obj_binder
                else:
                    binder = None
            elif method is d:
                # some non-method descriptor
                binder = _name_binder
            else:
                binder = None
            if binder is None:
                raise TypeError(
                    "'descriptor_bind' fails to auto-detect binding rule "
                    "of the given descriptor. Specify the rule by "
                    "'wirerope.wire.descriptor_bind.register'.")
            _descriptor_binder_cache[key] = binder
        else:
            binder = _descriptor_binder_cache[key]
        return binder


class Callable(object):
    """A wrapper object including more information of callables."""

    def __init__(self, f):
        self.wrapped_object = f
        self.is_function_type = type(self) is types.FunctionType  # noqa
        if self.is_descriptor:
            self.descriptor = Descriptor(f)
            f = getattr(f, self.descriptor.detect_function_attr_name())
        else:
            self.descriptor = None
        self.wrapped_callable = f
        self.is_wrapped_coroutine = getattr(f, '_is_coroutine', None)
        self.is_coroutine = self.is_wrapped_coroutine or \
            _inspect_iscoroutinefunction(f)

    @cached_property
    def signature(self):
        return inspect.signature(self.wrapped_callable)

    @cached_property
    def parameters(self):
        return list(self.signature.parameters.values())

    @property
    def first_parameter(self):
        return self.parameters[0] if self.parameters else None

    @cached_property
    def is_boundmethod(self):
        if self.is_function_type or self.is_builtin_property:
            return False
        new_bound = self.wrapped_object.__get__(object())
        try:
            if six.PY2:
                return new_bound is self.wrapped_object
            else:
                return type(new_bound) is type(self.wrapped_object)  # noqa
        except Exception:
            return False

    if six.PY2:
        @property
        def is_unboundmethod(self):
            return type(self.wrapped_object) is type(Callable.__init__)  # noqa

    @cached_property
    def is_descriptor(self):
        if self.is_boundmethod:
            return False
        is_descriptor = type(self.wrapped_object).__get__ \
            is not types.FunctionType.__get__  # noqa
        if six.PY2:
            is_descriptor = is_descriptor and \
                not (self.is_unboundmethod or self.is_boundmethod)
        return is_descriptor

    @cached_property
    def is_builtin_property(self):
        return issubclass(type(self.wrapped_object), property)

    @cached_property
    def is_property(self):
        return self.is_builtin_property or \
            (self.is_descriptor and self.descriptor.detect_property(
                _reagent, _Reagent))

    if six.PY34:
        @cached_property
        def is_barefunction(self):
            cc = self.wrapped_callable
            method_name = cc.__qualname__.split('<locals>.')[-1]
            if method_name == cc.__name__:
                return True
            return False
    else:
        @cached_property
        def is_barefunction(self):
            # im_class does not exist at this point
            if self.is_descriptor:
                return False
            if six.PY2:
                if self.is_unboundmethod:
                    return False
            return not (self.is_membermethod or self.is_classmethod)

    @cached_property
    def is_member(self):
        """Test given argument is a method or not.

        :rtype: bool

        :note: The test is partially based on the first parameter name.
            The test result might be wrong.
        """
        if six.PY34:
            if self.is_barefunction:
                return False
            if not self.is_descriptor:
                return True
        return self.first_parameter is not None \
            and self.first_parameter.name == 'self'

    @cached_property
    def is_membermethod(self):
        """Test given argument is a method or not.

        :rtype: bool

        :note: The test is partially based on the first parameter name.
            The test result might be wrong.
        """
        if self.is_boundmethod:
            return False
        if self.is_property:
            return False
        return self.is_member

    @cached_property
    def is_classmethod(self):
        """Test given argument is a classmethod or not.

        :rtype: bool

        :note: The test is partially based on the first parameter name.
            The test result might be wrong.
        """
        if isinstance(self.wrapped_object, classmethod):
            return True
        if six.PY34:
            if self.is_barefunction:
                return False
            if not self.is_descriptor:
                return False

        return self.first_parameter is not None \
            and self.first_parameter.name == 'cls'



import sys

from wirerope import Wire, WireRope, RopeCore

import pytest


class hybridmethod(object):

    def __init__(self, func):
        self.__func__ = func

    def __get__(self, obj, type=None):
        bound = obj if obj is not None else type
        return self.__func__.__get__(bound, type)


class hybridproperty(object):

    def __init__(self, func):
        self.__func__ = func

    def __get__(self, obj, type=None):
        bound = obj if obj is not None else type
        return self.__func__.__get__(bound, type)()


class awfuldescriptor(object):

    def __init__(self, func):
        self.__func__ = func

    def __get__(self, obj, type=None):
        return object()


def test_hybridmethod():

    class X(object):

        @hybridmethod
        def f(self):
            return self

    x = X()

    assert X.f() is X
    assert x.f() is x


def test_hybridproperty():

    class X(object):

        @hybridproperty
        def p(self):
            return self

    x = X()
    assert X.p is X
    assert x.p is x


def test_default_wire():
    class TestWire(Wire):
        pass

    rope = WireRope(TestWire, wraps=True)

    @rope
    def function(v):
        '''function'''
        return v

    class X(object):
        def baremethod(self, v):
            '''baremethod'''
            return (self, v)

        @rope
        def method(self, v):
            '''method'''
            return (self, v)

        @rope
        @classmethod
        def cmethod(cls, v):
            '''cmethod'''
            return (cls, v)

        @rope
        @staticmethod
        def smethod(v):
            '''smethod'''
            return (None, v)

        @rope
        @hybridmethod
        def hmethod(self_or_cls, v):
            '''hmethod'''
            return (self_or_cls, v)

        @rope
        @property
        def property(self):
            '''property'''
            return (self, )

        @rope
        @hybridproperty
        def hproperty(self_or_cls):
            '''hproperty'''
            return (self_or_cls,)

    assert isinstance(function, RopeCore)
    assert isinstance(X.method, Wire)  # triggered descriptor
    assert isinstance(X.cmethod, Wire)  # triggered descriptor
    assert isinstance(X.smethod, Wire)  # triggered descriptor
    assert isinstance(X.hmethod, Wire)  # triggered descriptor
    # assert isinstance(X.property, Wire)  # triggered descriptor

    x = X()

    boundmethod = rope(x.baremethod)

    assert x.method._owner == x
    assert x.cmethod._owner == X
    assert x.smethod._owner == X
    assert x.hmethod._owner == x
    assert X.cmethod._owner == X
    assert X.smethod._owner == X
    assert X.hmethod._owner == X
    assert x.method._bound_objects == (x,)
    assert x.cmethod._bound_objects == (X,)
    assert x.smethod._bound_objects == ()
    assert x.hmethod._bound_objects == (x,)
    assert X.cmethod._bound_objects == (X,)
    assert X.smethod._bound_objects == ()
    assert X.hmethod._bound_objects == (X,)

    assert not callable(function)
    assert not callable(boundmethod)
    assert not callable(x.method)
    assert not callable(x.cmethod)
    assert not callable(x.smethod)
    assert not callable(x.hmethod)
    assert not callable(x.property)
    assert not callable(x.hproperty)

    assert function.__func__(1) == 1
    assert boundmethod.__func__(1) == (x, 1)
    assert x.method.__func__(2) == (x, 2)
    assert X.method.__func__(x, 7) == (x, 7)
    assert x.cmethod.__func__(3) == (X, 3)
    assert X.cmethod.__func__(4) == (X, 4)
    assert x.smethod.__func__(5) == (None, 5)
    assert X.smethod.__func__(6) == (None, 6)
    assert x.hmethod.__func__(8) == (x, 8)
    assert X.hmethod.__func__(9) == (X, 9)
    assert x.property == (x, )
    assert x.hproperty == (x, )
    assert X.hproperty == (X, )

    assert function.__doc__ == 'function'
    assert boundmethod.__doc__ == 'baremethod'
    assert x.method.__doc__ == 'method'
    assert X.method.__doc__ == 'method'
    assert x.cmethod.__doc__ == 'cmethod'
    assert X.cmethod.__doc__ == 'cmethod'
    assert x.smethod.__doc__ == 'smethod'
    assert X.smethod.__doc__ == 'smethod'
    assert x.hmethod.__doc__ == 'hmethod'
    assert X.hmethod.__doc__ == 'hmethod'
    assert X.property.__doc__ == 'property'


def test_wire_super():
    if sys.version_info < (3, 6, 0):
        pytest.skip(
            "super() support requires __set_name__, which is not available"
            " before Python 3.6"
        )

    class CallableWire(Wire):
        def __call__(self, *args, **kwargs):
            return self.__func__(*args, **kwargs)

    rope = WireRope(CallableWire, wraps=True)

    class X(object):
        def baremethod(self, v):
            return "baremethod" + v

        @rope
        def method(self, v):
            return "method" + v

        @rope
        @classmethod
        def cmethod(cls, v):
            return "cmethod" + v

        @rope
        @staticmethod
        def smethod(v):
            return "smethod" + v

        @rope
        @hybridmethod
        def hmethod(self_or_cls, v):
            return "hmethod" + v

        @rope
        @property
        def property(self):
            return "property"

        @rope
        @hybridproperty
        def hproperty(self_or_cls):
            return "hproperty"

    class Y(X):
        def baremethod(self, v):
            return super(Y, self).baremethod(v) + "Y"

        @rope
        def method(self, v):
            return super(Y, self).method(v) + "Y"

        @rope
        @classmethod
        def cmethod(cls, v):
            return super(Y, cls).cmethod(v) + "Y"

        @rope
        @staticmethod
        def smethod(v):
            return X.smethod(v) + "Y"

        @rope
        @hybridmethod
        def hmethod(self_or_cls, v):
            # return super(Y, self_or_cls).hmethod(v) + "Y"
            return "hmethod Y"

        @rope
        @property
        def property(self):
            # return super(Y, self).property + "Y"
            return "propertyY"

        @rope
        @hybridproperty
        def hproperty(self_or_cls):
            # return super(Y, self_or_cls).hproperty + "Y"
            return "hpropertyY"

    for obj, suffix in [(X(), ""), (Y(), "Y")]:
        assert obj.baremethod(" ") == "baremethod " + suffix
        assert obj.method(" ") == "method " + suffix
        assert obj.cmethod(" ") == "cmethod " + suffix
        assert obj.smethod(" ") == "smethod " + suffix
        assert obj.hmethod(" ") == "hmethod " + suffix
        assert obj.property == "property" + suffix
        assert obj.hproperty == "hproperty" + suffix


def test_callable_wire():

    class CallableWire(Wire):

        def __call__(self, *args, **kwargs):
            return self.__func__(*args, **kwargs)

    rope = WireRope(CallableWire)

    @rope
    def f(v):
        return v

    class X(object):

        @rope
        def g(self, v):
            return v

        @rope
        @classmethod
        def h(cls, v):
            return v

        @rope
        @staticmethod
        def k(v):
            return v

        @rope
        @property
        def p(self):
            return 42

    x = X()

    assert callable(f)
    assert callable(x.g)
    assert callable(x.h)
    assert callable(x.k)
    assert not callable(x.p)

    assert f(1) == 1
    assert x.g(2) == 2
    assert X.g(x, 3) == 3
    assert x.h(4) == 4
    assert X.k(5) == 5
    assert x.p == 42


def test_wire():
    class TestWire(Wire):
        x = 7

        def y(self):
            if self._bound_objects:
                return self._bound_objects[0].v
            else:
                return None

    test_rope = WireRope(TestWire)

    @test_rope
    def f():
        return 'a'

    assert f.x == 7
    assert f.y() is None

    class A(object):

        def __init__(self, v):
            self.v = v

        @test_rope
        def f(self):
            return self.v

    a = A(10)
    b = A(20)

    assert a.f.x == 7
    assert a.f.y() == 10
    assert b.f.y() == 20
    assert not callable(a.f)


def test_unwirable():
    rope = WireRope(Wire)

    @rope
    def function(v):
        return v

    class X(object):

        @rope
        @awfuldescriptor
        def messed_up(self, v):
            return (self, v)

    with pytest.raises(TypeError):
        X.messed_up



import setuptools
from pkg_resources import get_distribution

try:
    get_distribution("setuptools>=39.2.0")
except Exception as e:
    raise AssertionError(
        "Please upgrade setuptools by `pip install -U setuptools`: {}".format(
            e
        )
    )

setuptools.setup()



""":mod:`wirerope.wire` --- end-point instant for each bound method
===================================================================
"""
import six
import types
from .callable import Descriptor
from ._compat import functools

__all__ = 'Wire',


@functools.singledispatch
def descriptor_bind(descriptor, obj, type_):
    binder = Descriptor(descriptor).detect_binder(obj, type_)
    return binder(descriptor, obj, type_)


@descriptor_bind.register(types.FunctionType)
def descriptor_bind_function(descriptor, obj, type):
    return obj, obj


class Wire(object):
    """The core data object for each function for bound method.

    Inherit this class to implement your own Wire classes.

    - For normal functions, each function is directly wrapped by **Wire**.
    - For any methods or descriptors (including classmethod, staticmethod),
      each one is wrapped by :class:`wirerope.rope.MethodRopeMixin`
      and it creates **Wire** object for each bound object.
    """

    __slots__ = (
        '_rope', '_callable', '_binding', '__func__', '_owner',
        '_bound_objects')

    def __init__(self, rope, owner, binding):
        self._rope = rope
        self._callable = rope.callable
        self._owner = owner
        self._binding = binding
        if binding:
            func = self._callable.wrapped_object.__get__
            if self._callable.is_property:
                wrapped = functools.partial(func, *binding)
                if six.PY2:
                    # functools.wraps requires those attributes but
                    # py2 functools.partial doesn't have them
                    wrapped.__module__ = owner.__module__
                    wrapped.__name__ = func.__name__
                self.__func__ = wrapped
            else:
                self.__func__ = func(*binding)
        else:
            self.__func__ = self._callable.wrapped_object
        if self._binding is None:
            self._bound_objects = ()
        else:
            _, binder = descriptor_bind(
                self._callable.wrapped_object, *self._binding)
            if binder is not None:
                self._bound_objects = (binder, )
            else:
                self._bound_objects = ()
        assert callable(self.__func__), self.__func__
        if rope._wrapped:
            functools.wraps(self.__func__)(self)

    def _on_property(self):
        return self.__func__()
